# The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.model import ModelSQL, ModelView, fields, Exclude
from sql.conditionals import Coalesce
from sql.operators import Equal as SqlEqual
from trytond.pool import Pool
from io import open
import oyaml as yaml
import os

__all__ = ['EdocumentMessage', 'EdocumentTemplate']

KNOWN_EXTENSIONS = ['.txt', '.edi', '.pla']


class EdocumentTemplate(ModelView, ModelSQL):
    """Electronic Document Template"""
    __name__ = 'edocument.template'
    rec_name = fields.Function(fields.Char('Record Name'), 'get_rec_name')

    message_type = fields.Selection([('DESADV', 'DESADV'),
        ('INVOIC', 'INVOIC')], 'Document', required=True)
    party = fields.Many2One('party.party', 'Party')
    template = fields.Text('Template')
    encoding = fields.Selection([
        ('utf-8', 'Utf-8'),
        ('cp1252', 'ANSI'),
        ], 'Encoding', required=True)

    @classmethod
    def default_encoding(cls):
        return 'utf-8'

    @classmethod
    def __setup__(cls):
        super(EdocumentTemplate, cls).__setup__()

        t = cls.__table__()
        cls._sql_constraints = [
            ('document_party_exclude', Exclude(t,
                (t.message_type, SqlEqual),
                (Coalesce(t.party, -1), SqlEqual)),
             'Combination of message type and party must be unique.')]

    def get_rec_name(self, name):
        return self.message_type + ' ' + self.party.name


class EdocumentMessage(ModelView, ModelSQL):
    """EDIFACT message"""
    __name__ = 'edocument.message'
    rec_name = fields.Function(fields.Char('Record Name'), 'get_rec_name')

    code = fields.Char('Code')
    message = fields.Text('Message')
    sequence = fields.Many2One('ir.sequence', 'Sequence')
    origin = fields.Reference('Origin', selection='get_origin')

    def get_rec_name(self, name):
        return 'EW' + '{0:0>6}'.format(self.code)

    @classmethod
    def get_origin(cls):
        IrModel = Pool().get('ir.model')
        models = cls._get_origin()
        models = IrModel.search([
                ('model', 'in', models),
                ])
        return [(None, '')] + [(m.model, m.name) for m in models]

    @classmethod
    def _get_origin(cls):
        return []

    @classmethod
    def create(cls, vlist):
        pool = Pool()
        Configuration = pool.get('edocument.configuration')
        Sequence = Pool().get('ir.sequence')
        config = Configuration(1)

        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if config.edocument_sequence:
                values['code'] = Sequence.get_id(config.edocument_sequence.id)
        return super(EdocumentMessage, cls).create(vlist)


class EdocumentMixin():
    messages = fields.One2Many('edocument.message', 'origin', 'Message')
    edocument_processed = fields.Function(
        fields.Boolean('Processed'), 'get_edocument_processed',
        searcher='search_edocument_processed')

    edi_message = fields.Function(
        fields.Char('EDI Code'), 'get_edi_message')

    def get_edocument_processed(self, name=None):
        return len(self.messages) > 0

    @classmethod
    def search_edocument_processed(cls, name, clause):
        exist = {
            True: [('messages', '=', None)],
            False: [('messages', '!=', None)]}
        return exist[bool(clause[1] == '=') ^ bool(clause[2])]

    def get_edi_message(self, name=None):
        return self.messages[-1].rec_name if self.messages else ''


class EdocumentImportMixin(EdocumentMixin):

    @classmethod
    def _get_template_name(cls):
        pass

    @classmethod
    def create_from_edi(cls, edi_info, template):
        pass

    @classmethod
    def postprocess(cls, objects):
        pass

    @classmethod
    def _source_uri(cls):
        pool = Pool()
        Configuration = pool.get('edocument.configuration')
        config = Configuration(0)
        return config.import_path

    @classmethod
    def _error_uri(cls):
        pool = Pool()
        Configuration = pool.get('edocument.configuration')
        config = Configuration(0)
        return config.error_path

    @classmethod
    def _template(cls):
        template_path = cls._get_template_name()
        with open(template_path, encoding='utf-8') as fp:
            return yaml.load(fp.read())

    @classmethod
    def _get_document_manager(cls):
        return EdocumentFileManager()

    @classmethod
    def _create_objects_from_edi(
            cls, source_uri, error_uri, template=None, target_model=None):
        """
        Get objects from edi files
        """
        pool = Pool()
        if not target_model:
            target_model = cls.__name__
        Object = pool.get(target_model)
        Message = pool.get('edocument.message')
        if not template:
            template = cls._template()
        objects = []
        messages = []
        imported = []
        manager = cls._get_document_manager()
        for name, info in manager._get_messages(source_uri, error_uri).items():
            created, errors = cls.create_from_edi(info, template)
            if created:
                message = Message(code=info)
                messages.append(message)
                imported.append(name)
                for value in created:
                    if 'id' in value:
                        new_object = Object(value['id'])
                    else:
                        new_object = Object()
                    for k, v in value.items():
                        setattr(new_object, k, v)
                    new_object.message = message
                    message.origin = new_object
                    objects.append(new_object)
            if imported:
                manager._handle_imported(imported)
            if errors:
                manager._handle_errors(name, errors, error_uri)
        Message.save(messages)
        Object.save(objects)
        return objects

    @classmethod
    def get_objects_from_edi(cls, target_model=None):
        '''Get objects from edi'''
        results = cls._create_objects_from_edi(
            cls._source_uri(),
            cls._error_uri(),
            target_model=target_model)
        cls.postprocess(results)
        return results


class EdocumentFileManager():
    ''' Imports EDI objects from a folder '''

    @classmethod
    def _read_file(cls, filename):
        with open(filename, encoding='utf-8') as file:
            return file.read()

    @classmethod
    def _get_messages(cls, uri, error_uri):
        files = [os.path.join(uri, file) for file in
                 os.listdir(uri) if os.path.isfile(os.path.join(uri, file))]

        return {fname: cls._read_file(fname) for fname in files
            if fname[-4:].lower() in KNOWN_EXTENSIONS}

    @classmethod
    def _handle_errors(cls, fname, errors, error_uri):
        error_fname = os.path.join(
            error_uri, 'error_{}.EDI'.format(
                os.path.splitext(os.path.basename(fname))[0]))
        with open(error_fname, 'w') as fp:
            fp.write('\n'.join(errors))

    @classmethod
    def _handle_imported(cls, imported):
        for file in imported:
            os.remove(file)
